from enum import IntEnum


__all__ = ['Filter', 'Event']


class FilterType(IntEnum):
    INCOMING = 1 << 0
    OUTGOING = 1 << 1
    ALL = INCOMING | OUTGOING

    @staticmethod
    def from_str(filter_type):
        if filter_type == 'i':
            return FilterType.INCOMING
        elif filter_type == 'o':
            return FilterType.OUTGOING
        else:
            return None


class Filter(dict):
    def __init__(self, filter_type: FilterType, connection_name_pattern: str):
        dict.__init__(self, type=filter_type, connectionNamePattern=connection_name_pattern)

    @staticmethod
    def from_str(filter_str):
        """
        Parse the filter string.

        @return Filter object if successful, None otherwise.
        """
        filter = filter_str.split(':')
        if len(filter) not in range(1, 3):  # 1 or 2 filter arguments
            return None
        filter_type = filter[0]
        if filter_type != 'i' and filter_type != 'o':
            return None

        connection_name_pattern = filter[1] if len(filter) > 1 else '*'       
        
        return Filter(FilterType.from_str(filter_type), connection_name_pattern)

    @staticmethod
    def default():
        return Filter(FilterType.ALL, '*')
    

class Event:
    def __init__(self, symbol):
        self.symbol = symbol
        self.data = None



from os import path
from collections import OrderedDict

import frida

from ..lib.types import Event
import datetime

class Agent:
    def __init__(self, filter, should_parse, session, reactor, print_timestamp=False):
        """
        Initialize the Frida agent
        """
        self._pending_events = OrderedDict() # A map of stacks, each stack holding events for that particular timestamp
        self._filter = filter
        self._should_parse = should_parse
        self._print_timestamp = print_timestamp
        self._script_path = path.join(path.abspath(path.dirname(__file__)), '../../_agent.js')
        with open(self._script_path) as src_f:
            script_src = src_f.read()
        self._script = session.create_script(script_src)
        self._reactor = reactor
        self._agent = None
        

    def start_hooking(self, ui):
        def on_message(message, data):
            self._reactor.schedule(lambda: self._on_message(message, data, ui))

        self._script.on('message', on_message)
        self._script.load()
        ui._update_status("Installing hooks...")
        self._agent = self._script.exports
        self._agent.install_hooks(self._filter, self._should_parse)

    def _on_message(self, message, data, ui):
        if message['type'] == 'error':
            click.secho(message['stack'], fg='red');
        mtype = message['payload']['type']

        if mtype == 'agent:hooks_installed':
            ui._update_status("Hooks installed, intercepting messages...")
        elif mtype == 'agent:trace:symbol':
            symbol = message['payload']['message']['symbol']
            timestamp = message['payload']['message']['timestamp']
            if timestamp in self._pending_events:
                self._pending_events[timestamp].append(Event(symbol)) 
                #print(f"Update {timestamp}")
            else:
                self._pending_events.update({ timestamp: [Event(symbol)] })
                #print(f"Add {timestamp}")
        elif mtype == 'agent:trace:data':
            timestamp = message['payload']['message']['timestamp']
            data = message['payload']['message']['data']
            self._pending_events[timestamp][-1].data = data
        else:
            ui._print(f"Unhandled message {message}")

        self.flush_pending_events()

    def flush_pending_events(self):
        """Flush pending events that are ready, i.e. have received both its symbol and data"""
        for ts, events_stack in list(self._pending_events.items()):
            while len(events_stack) > 0:
                last_event = events_stack[-1]  # Peek
                if last_event.data == None:
                    return
                #print(f"Pop {ts}")
                print('\n' + '-' * 60)
                if (self._print_timestamp):
                    date_time = datetime.datetime.fromtimestamp((ts/1000))
                    print(f"{date_time}")
                print(f"{last_event.symbol}\n{last_event.data['conn']}\n{last_event.data['message']}")
                print('-' * 60 + '\n')
                events_stack.pop()
            del self._pending_events[ts]



from frida_tools.application import ConsoleApplication
from frida_tools.tracer import UI, OutputFile

from ..utils.agent import Agent
from ..lib.types import Filter


class XPCSpyApplication(ConsoleApplication, UI):
    def _usage(self):
        return "usage: %prog [options] target"

    def _needs_target(self):
        return True

    def _add_options(self, parser):
        parser.add_option('-t', '--filter', 
                        help="Filter by message direction and service name. 'i' denotes incoming and 'o' denotes outgoing. Service name can include the wildcard character '*'. For exmaple 'i:com.apple.*' or 'o:com.apple.apsd'.",
                        metavar='FILTER', type='string')
        parser.add_option('-r', '--parse',
                        help="Parse XPC dictionary keys that include either `bplist00` or `bplist16` data.",
                        metavar='SHOULD_PARSE', action='store_true') 
        # parser.add_option('-o', '--output', help="dump output to file OUTPUT", metavar='OUTPUT', type='string')
        parser.add_option('-d', '--print-date',
                        help='Print a current timestamp before every XPC message', 
                        action='store_true', default=False)

    def _initialize(self, parser, options, args):
        if options.filter:
            filter = Filter.from_str(options.filter)
            if filter is None:
                self._update_status("Error: invalid filter string")
                self._exit(1)
            self._filter = filter
        else:
            self._filter = Filter.default()
        self._should_parse = options.parse or False
        self._print_timestamp = options.print_date or False 

    def _start(self):
        agent = Agent(self._filter, self._should_parse, self._session, self._reactor, self._print_timestamp)
        agent.start_hooking(self)


def main():
    app = XPCSpyApplication()
    app.run()



from setuptools import setup, find_packages
import os

from xpcspy.__init__ import __version__


cwd = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(cwd, 'requirements.txt'), 'r') as f:
    requirements = f.readlines()

setup(
    name='xpcspy',
    description="XPC message interception and more",
    license='Apache-2.0',
    author='hot3eed',
    author_email='hot3eed@gmail.com',
    url='https://github.com/hot3eed/xpcspy',
    keywords=['macos', 'ios', 'xnu', 'xpc', 'frida'],
    version=__version__,
    packages=find_packages(),
    install_requires=requirements,
    package_data={
        'xpcspy': [os.path.join(cwd, './_agent.js')]
        },
    entry_points={
        'console_scripts': [
                'xpcspy=xpcspy.console.cli:main'
            ]
        }
)

